#!/usr/bin/env python3
# -*- coding: utf-8 -*-
u"""
Created by ygidtu@gmail.com at 2022.05.11

The scripts contain the object to handle all the pysam, pybigwig related reading
"""
import os
from typing import Optional

import numpy as np
import pyBigWig
import pysam
from loguru import logger

from sashimi.base.GenomicLoci import GenomicLoci


def __opposite_strand__(strand: str) -> str:
    u"""
    replace to opposite of current strand information
    :param strand: strand, one of ['+', '-', '*']
    :return: an opposite of strand
    """
    assert strand in ["+", "-", "*"], "Unknown strand information was found."
    if strand == "*":
        return "*"
    elif strand == "+":
        return "-"
    else:
        return "+"


def __get_strand__(read: pysam.AlignedSegment, library: str) -> str:
    u"""
    Determine the strand of for each read.
    :param read: a pysam.AlignedSegment from the bam file.
    :param library: the method for preparing of the library,
        value should be one of [frf: "fr-firststrand", frs: "fr-secondstrand", fru: "fr-unstrand"]
    :return:
    """
    assert library in ["frf", "frs", "fru"], "Can't recognize the definition of library."

    # return '+' strand for all unstrand library.
    if library == "fru":
        return "+"

    # Only guessing the strand based on fr-firststrand rule.
    if read.is_paired:
        if read.is_read1 and read.is_reverse:
            current_strand = "+"
        elif read.is_read2 and not read.is_reverse:
            current_strand = "+"
        else:
            current_strand = "-"
    else:
        current_strand = "+" if read.is_reverse else "-"
        """
        TODO, add a more elegant way to solve some single-end strand specific library.
        """
        if library == "frf":
            # here for 3' tag rna seq
            current_strand = __opposite_strand__(current_strand)

    if library == "frf":
        return current_strand
    else:
        return __opposite_strand__(current_strand)


class Reader(object):

    @classmethod
    def __modify_chrom__(cls, region: GenomicLoci, reader, parser=None):
        if not region.chromosome.startswith("chr"):
            logger.info("Guess need 'chr'")

            if parser:
                iter_ = reader.fetch(
                    "chr" + region.chromosome if region.chromosome != "MT" else "chrM",
                    region.start, region.end, parser=parser
                )
            else:
                iter_ = reader.fetch(
                    "chr" + region.chromosome if region.chromosome != "MT" else "chrM",
                    region.start, region.end
                )

        else:
            logger.info("Guess 'chr' is redundant")
            if parser:
                iter_ = reader.fetch(
                    region.chromosome.replace("chr", "") if region.chromosome != "chrM" else "MT",
                    region.start, region.end, parser=parser
                )
            else:
                iter_ = reader.fetch(
                    region.chromosome.replace("chr", "") if region.chromosome != "chrM" else "MT",
                    region.start, region.end
                )

        return iter_

    @classmethod
    def read_bam(cls, path: str, region: GenomicLoci, library: str = "fru"):
        chrom = region.chromosome
        if not os.path.exists(path + ".bai"):
            logger.info(f"try to create index for {path}")
            pysam.index(path)

        with pysam.AlignmentFile(path, 'rb') as bam_file:
            try:
                relevant_reads = bam_file.fetch(reference=chrom, start=region.start, end=region.end)
            except ValueError as err:
                logger.warning(err)
                relevant_reads = cls.__modify_chrom__(region, bam_file)

            for read in relevant_reads:
                yield read, __get_strand__(read, library=library)

    @classmethod
    def read_gtf(cls, path: str, region: GenomicLoci, bed: bool = False):
        with pysam.TabixFile(path) as r:
            try:
                iter_ = r.fetch(
                    region.chromosome,
                    region.start,
                    region.end,
                    parser=pysam.asGTF() if not bed else pysam.asBed())
            except ValueError:
                try:
                    iter_ = cls.__modify_chrom__(region, r, parser=pysam.asGTF() if not bed else pysam.asBed())
                except ValueError as err:
                    logger.warning("please check the input region and gtf files")
                    logger.error(err)
                    raise err

            for record in iter_:
                yield record

    @classmethod
    def read_bigwig(cls, path: str, region: GenomicLoci) -> np.array:
        with pyBigWig.open(path) as r:
            try:
                return r.values(region.chromosome, region.start, region.end + 1)
            except RuntimeError as e:
                logger.warning(e)

                logger.info("may be caused by the mismatch of chromosome")
                if region.chromosome.startswith("chr"):
                    return r.values(region.chromosome.replace("chr", ""), region.start, region.end + 1)
                else:
                    return r.values("chr" + region.chromosome, region.start, region.end + 1)

    @classmethod
    def read_bigbed(cls, path: str, region: GenomicLoci):
        with pyBigWig.open(path) as r:
            if not r.isBigBed():
                logger.warning(f"{path} don't look like bigbed file.")
                yield from []
            else:
                try:
                    iter_ = r.entries(region.chromosome, region.start, region.end + 1)
                except RuntimeError as e:
                    logger.warning(e)

                    logger.info("may be caused by the mismatch of chromosome")
                    if region.chromosome.startswith("chr"):
                        iter_ = r.entries(
                            region.chromosome.replace("chr", "") if region.chromosome != "chrM" else "MT",
                            region.start,
                            region.end + 1
                        )
                    else:
                        iter_ = r.entries(
                            "chr" + region.chromosome if region.chromosome != "MT" else "chrM",
                            region.start,
                            region.end + 1
                        )
                if not iter_:
                    iter_ = []
                yield from iter_

    @classmethod
    def read_depth(cls, path: str, region: Optional[GenomicLoci] = None):
        if not os.path.exists(path + ".tbi"):
            logger.info(f"create tbi index for {path}")
            pysam.tabix_index(
                path, seq_col=0,
                start_col=1, end_col=1,
                force=True, keep_original=True
            )

        with pysam.TabixFile(path) as r:
            if not region:
                iter_ = r.fetch()
            else:
                try:
                    iter_ = r.fetch(region.chromosome, region.start, region.end, parser=pysam.asTuple())
                except ValueError:
                    try:
                        iter_ = cls.__modify_chrom__(region, r, parser=pysam.asTuple())
                    except ValueError as err:
                        logger.warning("please check the input region and gtf files")
                        logger.error(err)
                        raise err

            for record in iter_:
                yield record

    @classmethod
    def read_hic(cls, path: str, region: GenomicLoci):
        from hicmatrix import HiCMatrix as hm
        os.environ['NUMEXPR_MAX_THREADS'] = '16'
        os.environ['NUMEXPR_NUM_THREADS'] = '8'
        hic = hm.hiCMatrix(path, f"{region.chromosome}:{region.start}-{region.end}")
        return hic


if __name__ == '__main__':
    pass



from sashimi.plot_func import *

from matplotlib import pyplot as plt


def test_density():
    from sashimi.file.Bam import Bam
    fig, ax = plt.subplots()
    region = GenomicLoci("chr1", 1270656, 1284730, "+")
    bam = Bam.create("../example/bams/1.bam")
    bam.load(region, log_trans="2")
    plot_density(ax, bam)
    plt.savefig("plot_density_bam.png")


def test_bw():
    from sashimi.file.Bigwig import Bigwig

    bw = Bigwig.create("../example/bws/1.bw", title="test")
    bw.load(GenomicLoci("chr1", 1270656, 1284730, "+"))
    fig, ax = plt.subplots()
    plot_density(ax, bw)
    plt.savefig("plot_density_bw.png")


def test_ref():
    region = GenomicLoci("chr1", 1270656, 1284730, "+")
    fig, ax = plt.subplots()
    ref = Reference.create(
        "../example/example.sorted.gtf.gz",
        primary=True,
        add_domain=False)
    ref.load(region)

    ref.add_interval(
        interval="../example/PolyASite.chr1.atlas.clusters.2.0.GRCh38.96.bed.gz",
        label="PolyASite"
    )

    ref.add_interval(
        interval="../example/PolyASite.chr1.atlas.clusters.2.0.GRCh38.96.simple.bed.gz",
        label="PolyASite_simple"
    )

    plot_reference(ax, ref,
                   show_gene=True,
                   show_id=True,
                   plot_domain=True,
                   show_exon_id=True
                   )

    plt.savefig("plot_reference.pdf")


def test_depth():
    from sashimi.file.Depth import Depth
    region = GenomicLoci("chr1", 1270656, 1284730, "+")
    depth = Depth.create("../example/depth.bgz")
    depth.load(region)
    fig, ax = plt.subplots(nrows=len(depth))
    idx = 0
    for x, y in depth.items():
        plot_density(ax[idx], region=region, data=y, y_label=x,
                     theme="ticks_blank" if idx < len(depth) - 1 else "ticks")
        idx += 1

    plt.savefig("plot_density_depth.png")


def test_heatmap_and_line():
    from matplotlib import gridspec
    from sashimi.file.Bam import Bam

    region = GenomicLoci("chr1", 1270656, 1284730, "+")
    graph_coords = init_graph_coords(region)
    data = {}
    attrs = {}
    colors = ["red", "blue", "yellow", "green", "grey"]
    for i in range(1, 5):
        bam = Bam.create(f"../example/bams/{i}.bam")
        bam.load(region)
        data[str(i)] = bam.data
        attrs[str(i)] = {"color": colors[i]}

    gs = gridspec.GridSpec(1, 2, width_ratios=(.99, .01), wspace=0.01, hspace=.15)
    ax_var = plt.subplot(gs[:, 0]),
    cbar_ax = plt.subplot(gs[:, 1])
    plot_heatmap(ax_var[0], cbar_ax, data, do_scale=True, clustering=True)
    plt.savefig("plot_heatmap.png")

    fig, ax = plt.subplots()
    plot_line(ax, data, line_attrs=attrs, graph_coords=graph_coords)
    plt.savefig("plot_line.png")


def test_graph_coord():
    from matplotlib import gridspec
    region = GenomicLoci(chromosome="1", start=100, end=800, strand="+")
    exon_width = .5
    exons = [
        [150, 300], [320, 450],
        [460, 500], [470, 500],
        [470, 600], [700, 800],
    ]

    gs = gridspec.GridSpec(2, 1)
    ax = plt.subplot(gs[0, 0])
    for ind, exon in enumerate(exons):
        s, e, strand = exon[0], exon[1], "+"
        x = [s, e, e, s]
        x = [i - region.start for i in x]
        y = [
            ind - exon_width / 2, ind - exon_width / 2,
            ind + exon_width / 2, ind + exon_width / 2
        ]
        ax.fill(x, y, 'k', lw=.5, zorder=20)

    ax.set_xbound(0, len(region))
    graph_coords = init_graph_coords(region, exons, intron_scale=.5)

    ax = plt.subplot(gs[1, 0])
    for ind, exon in enumerate(exons):
        s, e, strand = exon[0] - region.start, exon[1] - region.start, "+"
        x = [graph_coords[s], graph_coords[e], graph_coords[e], graph_coords[s]]
        y = [
            ind - exon_width / 2, ind - exon_width / 2,
            ind + exon_width / 2, ind + exon_width / 2
        ]
        ax.fill(x, y, 'k', lw=.5, zorder=20)
    ax.set_xbound(0, len(region))
    plt.savefig("plot_graph_coords.png")


def test_set_x_ticks():
    region = GenomicLoci(chromosome="1", start=100, end=120, strand="+")
    fig, ax = plt.subplots()

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.tick_params(left=False)
    set_x_ticks(ax, region, font_size=10, sequence={110: "A", 106: "C"})
    plt.savefig("plot_x_ticks.png")


def test_igv_plot():
    from sashimi.file.ReadSegments import ReadSegment
    fig, ax = plt.subplots()
    region = GenomicLoci("chr1", 13362, 29900, "+")

    rs = ReadSegment.create("../example/bams/WASH7P.bam",
                            features={"m6a": "ma", "real_strand": "rs", "polya": "pa"})
    rs.load(region)
    plot_igv_like(ax, {'full': rs}, y_label="fl")
    plt.savefig("test_igv_plot.pdf")


def test_igv_plot2():
    from sashimi.file.ReadSegments import ReadSegment
    fig, ax = plt.subplots()
    region = GenomicLoci("chr1", 1270656, 1284730, "+")

    rs = ReadSegment.create("../example/bams/0.bam",
                            features={"m6a": "ma", "real_strand": "rs", "polya": "pa"})
    rs.load(region)
    plot_igv_like(ax, {'full': rs}, y_label="fl")
    plt.savefig("test_igv_plot.2.pdf")


def test_igv_plot3():
    from sashimi.file.ReadSegments import ReadSegment
    fig, ax = plt.subplots()
    # 1: 10024601 - 10038168
    region = GenomicLoci("1", 10024601, 10038168, "+")
    rs = ReadSegment.create("../example/test.bed.gz")
    rs.load(region)
    plot_igv_like(ax, {'full': rs}, y_label="fl")
    plt.savefig("test_igv_plot.3.pdf")


def test_hic_plot():
    from sashimi.file.HiCMatrixTrack import HiCTrack
    fig, ax = plt.subplots()
    region = GenomicLoci("X", 2500000, 2600000, "*")
    hic = HiCTrack.create(path="../example/Li_et_al_2015.h5", label="Li",
                          depth=30000,
                          trans="log2"
                          )
    hic.load(
        region=region
    )
    plot_hic(ax, {hic.label: hic}, y_label=hic.label)
    plt.savefig("test_hic.pdf")


def test_motif():
    fig, ax = plt.subplots()

    motif = Motif.create("../example/motif.bed.gz", GenomicLoci("chr1", 100, 105, "*"))
    motif.load(GenomicLoci("chr1", 100, 105, "*"))
    print(motif.data)

    plot_motif(ax, motif, width=0.8)
    plt.savefig("test_motif.pdf")


# test_igv_plot()
# test_igv_plot2()
# test_ref()
test_motif()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
This file contains the object to handle bam file related issues.

changelog:
    1. add library parameter for determining of read strand at 2022.4.28.

"""
import gzip
import os
from typing import Optional, Set

import numpy as np
import pysam
from loguru import logger

from sashimi.base.GenomicLoci import GenomicLoci
from sashimi.base.Junction import Junction
from sashimi.base.ReadDepth import ReadDepth
from sashimi.base.Readder import Reader
from sashimi.file.File import SingleCell


class Bam(SingleCell):
    def __init__(self,
                 path: str, label: str = "",
                 title: str = "", barcodes: Optional[Set[str]] = None,
                 barcode_tag: str = "CB", umi_tag: str = "UB",
                 library: str = "fru", density_by_strand: bool = False):
        u"""
        init this object
        :param label: the left axis label
        :param title: the default title to show in the upper-right of density plot
        :param barcodes: the path to barcodes,
                default: ../filtered_feature_bc_matrix/barcodes.tsv.gz of bam file according to 10X Genomics
        :param barcode_tag: the cell barcode tag, default is CB according to 10X Genomics
        :param umi_tag: the UMI barcode tag, default is UB according to 10X Genomics
        :param library: library for determining of read strand.
        :param density_by_strand: whether to draw density plot in strand-specific manner.
        """
        super().__init__(path, barcodes, barcode_tag, umi_tag)
        self.title = title
        self.label = label if label else os.path.basename(path).replace(".bam", "")
        self.library = library
        self.density_by_strand = density_by_strand

    @classmethod
    def create(cls,
               path: str,
               label: str = "",
               title: str = "",
               barcodes: Optional[Set[str]] = None,
               barcode_tag: str = "CB",
               umi_tag: str = "UB",
               library: str = "fru",
               density_by_strand: bool = False
               ):
        u"""

        :param path: the path to bam file
        :param label: the left axis label
        :param title: the default title to show in the upper-right of density plot
        :param barcodes: the path to barcodes,
                default: ../filtered_feature_bc_matrix/barcodes.tsv.gz of bam file according to 10X Genomics
        :param barcode_tag: the cell barcode tag, default is CB according to 10X Genomics
        :param umi_tag: the UMI barcode tag, default is UB according to 10X Genomics
        :param library: library for determining of read strand.
        :return:
        """

        if not os.path.exists(path + ".bai"):
            pysam.index(path)

        barcode = barcodes
        path = os.path.abspath(path)
        if not barcodes:
            barcode = set()
            barcodes = os.path.join(os.path.dirname(path), "filtered_feature_bc_matrix/barcodes.tsv.gz")

            if os.path.exists(barcodes):
                with gzip.open(barcodes, "rt") as r:
                    for line in r:
                        barcode.add(line.strip())
        return cls(
            path=path,
            label=label,
            title=title,
            barcodes=barcode,
            barcode_tag=barcode_tag,
            umi_tag=umi_tag,
            library=library,
            density_by_strand=density_by_strand
        )

    def __hash__(self):
        return hash(self.label)

    def __str__(self) -> str:

        temp = []

        for x in [self.title, self.label, self.path]:
            if x is None or x == "":
                x = "None"
            temp.append(str(x))

        return "\t".join(temp)

    def to_csv(self) -> str:
        temp = []

        for x in [self.title, self.label, self.path]:
            if x is None or x == "":
                x = "None"
            if isinstance(x, list):
                x = ";".join(x)
            temp.append(str(x))

        return ",".join(temp)

    def load(self,
             region: GenomicLoci,
             threshold: int = 0,
             reads1: Optional[bool] = None,
             required_strand: Optional[str] = None,
             **kwargs
             ):
        """
            determine_depth determines the coverage at each base between start_coord and end_coord, inclusive.

            bam_file_path is the path to the bam file used to \
            determine the depth and junctions on chromosome between start_coord and end_coord

        return values:
            depth_vector,
            which is a Numpy array which contains the coverage at each base position between start_coord and end_coord

            spanned_junctions, which is a dictionary containing the junctions supported by reads.
            The keys in spanned_junctions are the
                names of the junctions, with the format chromosome:lowerBasePosition-higherBasePosition
        :param region: GenomicLoci object including the region for calculating coverage
        :param threshold: minimums counts of the given splice junction for visualization
        :param reads1: None -> all reads, True -> only R1 kept; False -> only R2 kept
        :param required_strand: None -> all reads, else reads on specific strand
        """
        self.region = region
        filtered_junctions = {}

        spanned_junctions = kwargs.get("junctions", {})
        included_junctions = kwargs.get("included_junctions", {})
        remove_duplicate_umi = kwargs.get("remove_duplicate_umi", False)
        spanned_junctions_plus = dict()
        spanned_junctions_minus = dict()
        plus, minus = np.zeros(len(region), dtype="f"), np.zeros(len(region), dtype="f")
        site_plus, site_minus = np.zeros(len(region), dtype="f"), np.zeros(len(region), dtype="f")

        umis = {}

        try:
            for read, strand in Reader.read_bam(path=self.path, region=region, library=self.library):
                # make sure that the read can be used
                cigar_string = read.cigartuples

                # each read must have a cigar string
                if cigar_string is None:
                    continue

                # select R1 or R2
                if reads1 is True and not read.is_read1:
                    continue

                if reads1 is False and not read.is_read2:
                    continue

                # filter reads by 10x barcodes
                # @20220924, add `not` before has_barcode and skip these reads without umi tag.
                if self.barcodes:
                    if not read.has_tag(self.barcode_tag) or not self.has_barcode(read.get_tag(self.barcode_tag)):
                        continue

                    if remove_duplicate_umi:
                        barcode = read.get_tag(self.barcode_tag)
                        if barcode not in umis.keys():
                            umis[barcode] = {}

                        # filter reads with duplicate umi by barcode
                        if read.has_tag(self.umi_tag):
                            umi = read.get_tag(self.umi_tag)

                            if umi in umis[barcode].keys() and umis[barcode][umi] != hash(read.query_name):
                                continue

                            if len(umis[barcode]) == 0:
                                umis[barcode][umi] = hash(read.query_name)
                        else:
                            continue

                start = read.reference_start
                if required_strand and strand != required_strand:
                    continue

                """
                M	BAM_CMATCH	0
                I	BAM_CINS	1
                D	BAM_CDEL	2
                N	BAM_CREF_SKIP	3
                S	BAM_CSOFT_CLIP	4
                H	BAM_CHARD_CLIP	5
                P	BAM_CPAD	6
                =	BAM_CEQUAL	7
                X	BAM_CDIFF	8
                B	BAM_CBACK	9
                """
                for cigar, length in cigar_string:
                    cur_start = start + 1
                    cur_end = start + length + 1

                    if cigar == 0:  # M
                        for i in range(length):
                            if region.start <= start + i + 1 <= region.end:
                                try:
                                    if strand == "+":
                                        plus[start + i + 1 - region.start] += 1
                                    elif strand == "-":
                                        minus[start + i + 1 - region.start] += 1
                                    else:
                                        pass
                                except IndexError as err:
                                    logger.info(region)
                                    logger.info(cigar_string)
                                    logger.info(start, i)
                                    exit(err)

                    # remove the deletion.
                    if cigar not in (1, 4, 5):  # I, S, H
                        start += length

                    if cigar == 3 and not kwargs.get("only_customized_junction"):  # N
                        try:
                            junction_name = Junction(region.chromosome, cur_start, cur_end, strand)

                            if junction_name not in spanned_junctions:
                                spanned_junctions[junction_name] = 0

                            spanned_junctions[junction_name] = spanned_junctions[junction_name] + 1
                        except ValueError as err:
                            logger.warning(err)
                            continue
                start = read.reference_start + 1 if read.reference_start + 1 > region.start else region.start
                end = read.reference_end + 1 if read.reference_end + 1 < region.end else region.end
                if strand == "+" and 0 <= start - region.start < len(plus):
                    site_plus[end - region.start] += 1
                elif strand == "-" and 0 <= end - region.start < len(minus):
                    site_minus[start - region.start] += 1

            for k, v in spanned_junctions.items():
                if included_junctions and str(k) not in included_junctions:
                    continue

                if v >= threshold:
                    filtered_junctions[k] = v

                    if k.strand == "-":
                        if k not in spanned_junctions_plus:
                            spanned_junctions_plus[k] = -1
                        else:
                            spanned_junctions_plus[k] += -1
                    elif k.strand == "+":
                        if k not in spanned_junctions_minus:
                            spanned_junctions_minus[k] = -1
                        else:
                            spanned_junctions_minus[k] += -1
        except IOError as err:
            logger.error('There is no .bam file at {0}'.format(self.path))
            logger.error(err)
        except ValueError as err:
            logger.error(self.path)
            logger.error(err)

        self.data = ReadDepth(
            plus if self.density_by_strand else plus + minus,
            junctions_dict=filtered_junctions,
            site_plus=site_plus,
            site_minus=site_minus,
            minus=minus if self.density_by_strand else None,
            junction_dict_plus=spanned_junctions_plus,
            junction_dict_minus=spanned_junctions_minus,
            strand_aware=False if self.library == "fru" else True)
        return self


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2019/1/17 3:27 PM
u"""
Created by Ran Zhou at 2019/1/17 3:27 PM
This dataset was curated from Singh et al.(10.1038/s41467-018-04112-z)

"""
u"""
# domain information which was downloaded from https://ebi-uniprot.github.io/ProtVista/userGuide.html

category	Type	Label	Description
DOMAIN_AND_SITES	domain	Domain	Position and type of each modular protein domain
DOMAIN_AND_SITES	repeat	Repeat	Positions of repeated sequence motifs or repeated domains
DOMAIN_AND_SITES	ca_bind	Calcium binding	Position(s) of calcium binding region(s) within the protein
DOMAIN_AND_SITES	zn_fing	Zinc finger	Position(s) and type(s) of zinc fingers within the protein
DOMAIN_AND_SITES	dna_bind	DNA binding	Position and type of a DNA-binding domain
DOMAIN_AND_SITES	np_bind	Nucleotide binding	Nucleotide phosphate binding region
DOMAIN_AND_SITES	region	Region	Region of interest in the sequence
DOMAIN_AND_SITES	coiled	Coiled coil	Positions of regions of coiled coil within the protein
DOMAIN_AND_SITES	motif	Motif	Short (up to 20 amino acids) sequence motif of biological interest
DOMAIN_AND_SITES	act_site	Active site	Amino acid(s) directly involved in the activity of an enzyme
DOMAIN_AND_SITES	metal	Metal binding	Binding site for a metal ion
DOMAIN_AND_SITES	binding	Binding site	Binding site for any chemical group (co-enzyme, prosthetic group, etc.)
DOMAIN_AND_SITES	site	Site	Any interesting single amino acid site on the sequence
MOLECULE_PROCESSING	init_met	Initiator methionine	Cleavage of the initiator methionine
MOLECULE_PROCESSING	signal	Signal	Sequence targeting proteins to the secretory pathway or periplasmic space
MOLECULE_PROCESSING	transit	Transit peptide	Extent of a transit peptide for organelle targeting
MOLECULE_PROCESSING	propep	Propeptide	Part of a protein that is cleaved during maturation or activation
MOLECULE_PROCESSING	chain	Chain	Extent of a polypeptide chain in the mature protein
MOLECULE_PROCESSING	peptide	Peptide	Extent of an active peptide in the mature protein
PTM	mod_res	Modified residue	Modified residues excluding lipids, glycans and protein cross-links
PTM	lipid	Lipidation	Covalently attached lipid group(s)
PTM	carbohyd	Glycosylation	Covalently attached glycan group(s)
PTM	disulfid	Disulfide bond	Cysteine residues participating in disulfide bonds
PTM	crosslnk	Cross-link	Residues participating in covalent linkage(s) between proteins
STRUCTURAL	helix	Helix	Helical regions within the experimentally determined protein structure
STRUCTURAL	turn	Turn	Turns within the experimentally determined protein structure
STRUCTURAL	strand	Beta strand	Beta strand regions within the experimentally determined protein structure
TOPOLOGY	topo_dom	Topological domain	Location of non-membrane regions of membrane-spanning proteins
TOPOLOGY	transmem	Transmembrane	Extent of a membrane-spanning region
TOPOLOGY	intramem	Intramembrane	Extent of a region located in a membrane without crossing it
"""

__VALID_DOMAIN_CATEGORY__ = {
    'DOMAIN_AND_SITES',
    'MOLECULE_PROCESSING',
    'PTM',
    'STRUCTURAL',
    'TOPOLOGY'
}

__DOMAINFILTER__ = {"active site", "domain", "signal peptide", "transmembrane region", "repeat", "zinc ﬁnger region",
                    "compositionally biased region", "DNA-binding region", "region of interest",
                    "lipid moiety-binding region", "short sequence motif", "calcium-binding region",
                    "nucleotide phosphate-binding region", "metal ion-binding site", "topological domain"}

__DNABIND__ = {'C2H2-type', 'PHD-type', 'C3H1-type', 'KRAB', 'Bromo', 'Chromo', 'DNA-binding', 'C4-type', 'CHCR',
               'A.T hook', 'bZIP', 'bHLH', 'CCHC-type', 'CHCH', 'Bromodomain-like', 'CH1', 'C6-type', 'A.T hook-like',
               'C4H2 - type', 'CHHC-type'}

__ACTIVE__ = {'active site', 'catalytic sites'}

__TRANSREGION__ = {'transmembrane region', 'ABC transmembrane type-1', 'ABC transporter', 'ABC transmembrane type-2'}

__PPI__ = {"WD", "ANK", "TPR", "LRR", "HEAT", "Sushi", "EF-hand", "ARM", "PDZ", "PH", "SH3", "RING-type",
           "LIM zinc-binding", "WW", "SH2", "BTB", "FERM", "CH", "Rod", "Coil 1A", "MH2", "WD40-like repeat",
           "t-SNARE coiled-coil homology", "Coil 1B", "Cbl-PTB", "Coil", "CARD", "SH2-like", "DED", "IRS-type PTB",
           "SP-RING-type", "EF-hand-like", "RING-CHtype", "v-SNARE coiled-coil homology", "Arm domain",
           "LIM protein-binding", "GYF", "PDZ domain-binding", "PDZD11-binding"}

__RNABIND__ = {"RRM", "SAM", "KH", "DRBM", "RBD", "Piwi", "PAZ", "S1 motif", "Pumilio", "THUMP"}

__ANNOTATION__ = {
    "PPI": __PPI__,
    "TMD": __TRANSREGION__,
    "activesite": __ACTIVE__,
    "dnabinding": __DNABIND__,
    "rnabinding": __RNABIND__
}



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
This file contains the object to handle depth file generated using samtools depth.
"""
import gzip
import os
from typing import List, Optional

import numpy as np

from sashimi.base.GenomicLoci import GenomicLoci
from sashimi.base.ReadDepth import ReadDepth
from sashimi.base.Readder import Reader
from sashimi.file.File import File


class Depth(File):
    u"""
    To generate depth file from bams
    1. samtools depth *.bam | bgzip > depth.bgz
    2. tabix -s 2 -e 2 depth.bgz
    """

    def __init__(self, path: str, label: List[str] = None, title: str = ""):
        super().__init__(path)
        self.label = label
        self.title = title
        self.region = None

    def __len__(self) -> int:
        return len(self.data)

    @classmethod
    def create(cls, path: str, label: List[str] = None, title: str = ""):
        u"""
        create depth obj
        """
        assert os.path.exists(path), f"{path} not exists."

        with gzip.open(path, "rt") as r:
            for line in r:
                lines = line.split()

                if label:
                    assert len(lines) - 2 == len(label), \
                        f"the number of columns [{len(lines) - 2}] must equal to the number of labels [{len(label)}]"
                else:
                    label = list(range(1, len(lines) - 1))
                break

        return cls(path, label, title)

    def load(self, region: GenomicLoci, required_sample: List[str] = None, log_trans: Optional[str] = None, **kwargs):
        u"""
        depth file generated by samtools depth

        samtools depth *.bam | bgzip > depth.gz && tabix indexed tabix -s 1 -b 2 -e 3 depth.gz
        :param region: the region to query
        :param required_sample: the sample to use
        :param log_trans: log transform
        :return:
        """
        self.region = region
        self.log_trans = log_trans
        if not required_sample:
            required_sample = self.label

        required_sample = [x for x in required_sample if x in self.label]
        depth_vector = {x: np.zeros(len(region), dtype='f') for x in required_sample}

        for row in Reader.read_depth(self.path, region):
            chrom, site = row[0], int(row[1]) - region.start

            if site < len(region):
                vals = {x: float(y) for x, y in zip(self.label, row[2:])}
                for x in depth_vector.keys():
                    depth_vector[x] += vals[x]

        self.data = {x: ReadDepth(y) for x, y in depth_vector.items()}

    def items(self):
        if self.data:
            for x, y in self.data.items():
                yield x, y


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
Created by ygidtu@gmail.com at 2019.12.06
"""
import copy
import faulthandler
import io
import logging
import os
from multiprocessing import Pool
from typing import Set

import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.backends.backend_pdf import FigureCanvasPdf

from sashimi.file.ATAC import ATAC
from sashimi.file.Bam import Bam
from sashimi.file.BedGraph import Bedgraph
from sashimi.file.Bigwig import Bigwig
from sashimi.file.Depth import Depth
from sashimi.file.Fasta import Fasta
from sashimi.file.Junction import load_custom_junction
from sashimi.file.Motif import Motif
from sashimi.plot_func import *

logging.getLogger('matplotlib.font_manager').setLevel(logging.ERROR)

faulthandler.enable()


def __load__(args):
    p, args, kwargs = args
    p.load(*args, **kwargs)
    return p


class PlotInfo(object):
    u"""
    this class is used to collect all the plot information.
    """

    def __init__(self, obj: File, category: str = "", type_: str = "", group: str = ""):
        u"""
        init this class
        :param obj: the input file object
        :param category: the input file category
        :param type_: the plot type
        :param group: group name of input file, only used to plots type_ == heatmap
        """
        self.obj = [obj]
        self.group = group
        self.type = type_
        self.category = [category]

    def __str__(self) -> str:
        return ";".join([obj.path for obj in self.obj])

    def __hash__(self) -> int:
        if self.type in ["heatmap", "line"]:
            return hash(self.group)
        return hash((tuple([hash(x) for x in self.obj]), self.group, self.type, tuple(self.category)))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return hash(self) != hash(other)

    def __add__(self, other):
        if self.category == ["heatmap", "line"]:
            self.obj[0] += other.obj
        else:
            self.obj += other.obj
        return self

    @property
    def is_single_cell(self):
        try:
            return self.obj[0].is_single_cell
        except AttributeError:
            return False

    @property
    def data(self) -> Dict[str, Union[ReadDepth, ReadSegment]]:
        data = {}
        for obj in self.obj:
            if isinstance(obj.data, dict):
                data.update(data)
            elif isinstance(obj, ReadSegment):
                data[obj.label] = obj
            else:
                data[obj.label] = obj.data
        return data

    def len(self, scale: Union[int, float] = .25) -> int:
        n = 0
        if not self.category:
            pass
        elif self.type == "site-plot" and self.category[0] == "bam":
            n += 2
        elif self.type == "igv":
            n += self.obj[0].len(scale / 8)
        else:
            n += 1
        return n

    def add(self, obj: File, category: str = "", type_: str = ""):
        u"""
        add new input file to specific group
        """
        assert type_ in ["heatmap", "line"], "only heatmap/line plot needs add"

        self.obj.append(obj)
        self.category.append(category)
        return self

    def load(self, region: GenomicLoci, n_jobs: int = 0, *args, **kwargs):
        if n_jobs <= 1:
            for obj in self.obj:
                obj.load(region=region, *args, **kwargs)
        else:
            with Pool(n_jobs) as p:
                kwargs_ = deepcopy(kwargs)
                kwargs_["region"] = region
                self.obj = p.map(__load__, [[o, args, kwargs_] for o in self.obj])

        if self.type == "density":
            if len(self.obj) > 1:
                data = self.obj[0].data
                for obj in self.obj[1:]:
                    data += obj.data

                for obj in self.obj:
                    obj.data = data

        return self


class Plot(object):
    u"""
    this class is the main framework of sashimi
    """

    def __init__(self):
        u"""
        init this class
        """
        self.__n_objs__ = 0
        self.region = None
        self.sites = {}
        self.focus = {}
        self.stroke = []
        self.events = None
        self.sequence = None
        self.reference = None
        self.graph_coords = None
        self.plots = []
        self.params = {}
        self.junctions = {}
        self.link = []

    @property
    def chrom(self) -> Optional[str]:
        if self.region:
            return self.region.chrom

    @property
    def start(self) -> Optional[int]:
        if self.region:
            return self.region.start

    @property
    def end(self) -> Optional[int]:
        if self.region:
            return self.region.end

    @property
    def strand(self) -> Optional[str]:
        if self.region:
            return self.region.strand

    @property
    def exons(self) -> Optional[List[List[int]]]:
        if self.reference:
            return self.reference.exons

    def set_region(self, chromosome: str = "",
                   start: int = 0, end: int = 0,
                   strand: str = "+",
                   region: Optional[GenomicLoci] = None):
        u"""
        change the plot region
        """

        if region:
            self.region = region
        elif chromosome and start and end:
            self.region = GenomicLoci(chromosome, start=start, end=end, strand=strand)
        logger.info(f"set region to {self.region}")
        return self

    def add_sites(self, sites):
        u"""
        highlight specific sites
        :param sites: string in 100,200 format or int
        :return:
        """
        assert self.region is not None, f"please set plot region first."
        if sites is not None:
            if isinstance(sites, str):
                sites = [int(x) for x in sites.split(",")]

                for s in sites:
                    if s not in self.sites.keys():
                        self.sites[s - self.start] = "blue"
                    else:
                        self.sites[s - self.start] = "red"
            elif isinstance(sites, int):
                if sites not in self.sites.keys():
                    self.sites[sites - self.start] = "blue"
                else:
                    self.sites[sites - self.start] = "red"
        return self

    def add_focus(self, focus: Optional[str], start: int = 0, end: int = 0):
        u"""
        set focus region
        :param focus: string in 100-200:300-400
        :param start: start site
        :param end: end site
        :return:
        """
        assert self.region is not None, f"please set plot region first."
        if focus:
            for site in focus.split(":"):
                site = sorted([int(x) - self.start for x in site.split("-")])
                if site[0] < 0:
                    site[0] = 0

                if site[-1] > len(self):
                    site[-1] = len(self)

                self.focus[site[0]] = max(site[1], self.focus.get(site[0], -1))

        if 0 < start < end:
            self.focus[start] = max(end, self.focus.get(start, -1))
        return self

    def add_stroke(
            self,
            stroke: Optional[str] = None,
            start: int = 0,
            end: int = 0,
            label: str = "",
            color: str = "black"
    ):
        u"""
        add stroke to plot
        :param stroke: stroke string in 100-200@red;300-400 format
        :param start: start site of stroke
        :param end: end site of stroke
        :param label: label of stroke
        :param color: color of stroke
        :return:
        """
        assert self.region is not None, f"please set plot region first."
        if stroke:
            self.stroke += Stroke.create(stroke, self.region)

        if 0 < start < end:
            self.stroke.append(Stroke(start - self.start, end - self.end, color, label))
        return self

    def add_links(
            self,
            link: Optional[str] = None,
            start: int = 0,
            end: int = 0,
            label: str = "",
            color: str = "blue"
    ):
        u"""
        add stroke to plot
        :param link: string in 100-200@red;300-400 format
        :param start: start site of stroke
        :param end: end site of stroke
        :param label: label of stroke
        :param color: color of stroke
        :return:
        """
        assert self.region is not None, f"please set plot region first."
        if link:
            self.link.append(Stroke.create(link, self.region, default_color=color))

        if 0 < start < end:
            self.link.append([Stroke(start - self.start, end - self.end, color, label)])
        return self

    def set_sequence(self, fasta: str):
        u"""
        set sequence info for
        :param fasta: path to indexed fasta file
        :return:
        """
        logger.info(f"fetch sequence from {fasta}")
        self.sequence = Fasta.create(fasta)
        return self

    def set_reference(self, gtf: str,
                      add_domain: bool = False,
                      local_domain: Optional[str] = False,
                      interval: Optional[str] = None,
                      interval_label: Optional[str] = None,
                      transcripts: Optional[List[str]] = None,
                      remove_empty_transcripts: bool = False,
                      choose_primary: bool = False,
                      color: Optional[str] = "black",

                      # transcripts related parameters
                      font_size: int = 5,
                      show_gene: bool = False,
                      show_id: bool = False,
                      reverse_minus: bool = False,
                      exon_width: float = .3,
                      show_exon_id: bool = False,
                      theme: str = "blank"
                      ):
        u"""
        add transcripts to this region
        :param gtf:
        :param add_domain:
        :param local_domain:
        :param interval:
        :param interval_label:
        :param font_size: the size of transcript id, name
        :param transcripts: the list of name or ids of transcripts to draw
        :param remove_empty_transcripts: whether to remove transcripts without any exons
        :param choose_primary: whether to choose primary transcript for each gene
        :param color: the color of exons
        :param show_gene: whether to show gene name/id
        :param show_id: show gene id or gene name
        :param reverse_minus: whether to remove strand of transcripts
        :param theme: the theme of transcript
        :param exon_width: the height of exons
        :param show_exon_id: whether to show exon id
        :return:
        """
        logger.info(f"set reference file to {gtf}")
        self.reference = Reference.create(
            gtf,
            add_domain=add_domain,
            add_local_domain=local_domain
        )

        if interval and interval_label:
            self.reference.add_interval(interval, interval_label)

        self.params["reference"] = {
            "transcripts": transcripts,
            "remove_empty_transcripts": remove_empty_transcripts,
            "choose_primary": choose_primary,
            "color": color,
            "font_size": font_size,
            "show_gene": show_gene,
            "show_id": show_id,
            "reverse_minus": reverse_minus,
            "exon_width": exon_width,
            "show_exon_id": show_exon_id,
            "theme": theme
        }

        return self

    def add_interval(self, interval: str, interval_label: str):
        u"""

        """
        assert self.reference is not None, "please set_reference first."
        self.reference.add_interval(interval, interval_label)
        return self

    def __init_input_file__(self, path: str,
                            category: str = "bam",
                            label: Union[str, List[str]] = "",
                            title: str = "",
                            barcode: str = "",
                            barcode_groups: Dict[str, Set[str]] = None,
                            barcode_tag: str = "BC",
                            umi_tag: str = "UB",
                            library: str = "fru",
                            features: Optional[dict] = None,
                            deletion_ignore: Optional[int] = True,
                            del_ratio_ignore: float = .5,
                            exon_focus: Optional[str] = None,
                            density_by_strand: bool = False,

                            # for hic plot
                            trans: Optional[str] = None,
                            depth: Optional[int] = 30000,

                            # for ATAC
                            size_factor=None,

                            log_trans: Optional[str] = None,
                            ):
        self.__n_objs__ += 1
        path = os.path.expanduser(path)
        if category == "bam":
            obj = Bam.create(
                path,
                label=label,
                title=title,
                barcodes=barcode_groups.get(barcode) if barcode_groups else None,
                barcode_tag=barcode_tag,
                umi_tag=umi_tag,
                library=library,
                density_by_strand=density_by_strand
            )
        elif category == "atac":
            obj = ATAC.create(
                path,
                label=label,
                title=title,
                barcode=barcode,
                barcode_groups=barcode_groups,
                size_factors=size_factor
            )
        elif category == "igv":
            obj = ReadSegment.create(
                path=path,
                label=label,
                library=library,
                features=features,
                deletion_ignore=deletion_ignore,
                del_ratio_ignore=del_ratio_ignore,
                exon_focus=exon_focus
            )
        elif category == "hic":
            obj = HiCTrack.create(
                path=path,
                label=label,
                trans=trans,
                depth=depth
            )
        elif category == "bigwig" or category == "bw":
            category = "bw"
            obj = Bigwig.create(path, label=label, title=title)
        elif category == "bedgraph" or category == "bg":
            category = "bg"
            obj = Bedgraph.create(path=path, label=label, title=title)
        elif category == "depth":
            obj = Depth.create(path, label=label, title=title)
        else:
            raise ValueError(
                f"the category should be one of [bam, bigwig, bw, depth, bedgraph, bg], instead of {category}")

        obj.log_trans = log_trans
        return obj, category

    def add_customized_junctions(self, path: str):
        if path and os.path.exists(path) and os.path.isfile(path):
            self.junctions = load_custom_junction(path)

    def add_density(self,
                    path: str,
                    category: str = "bam",
                    size_factor=None,

                    # file loading parameters
                    label: Union[str, List[str]] = "",
                    title: str = "",
                    barcode: str = "",
                    barcode_groups: Dict[str, Set[str]] = None,
                    barcode_tag: str = "BC",
                    umi_tag: str = "UB",
                    library: str = "fru",
                    density_by_strand: bool = False,

                    # plotting parameters
                    color="blue",
                    font_size: int = 8,
                    show_junction_number: bool = True,
                    junction_number_font_size: int = 5,
                    n_y_ticks: int = 4,
                    show_y_label: bool = True,
                    y_label: str = "",
                    theme: str = "ticks_blank",
                    log_trans: Optional[str] = None,

                    # site plot parameters
                    show_site_plot: bool = False,
                    strand_choice: Optional[str] = None,

                    only_customized_junction: bool = False
                    ):
        u"""
        add density object to plot
        :param path: the path to input file
        :param category: the input file type
        :param show_site_plot: draw the density distribution of reads from different strand
        :param label: the label of input file
        :param title: the title of input file
        :param size_factor
        :param barcode: key of barcode barcode_groups
        :param barcode_groups:
        :param barcode_tag: cell barcode tag
        :param umi_tag: umi barcode tag
        :param density_by_strand: whether to draw density plot in strand-specific manner.
        :param library: should be one of [frf: "fr-firststrand", frs:"fr-secondstrand", fru:"fr-unstrand"], default: fru
        :param font_size: the font size for ticks, y-axis label and title
        :param show_junction_number: whether to show the number of junctions
        :param n_y_ticks: number of y ticks
        :param junction_number_font_size:
        :param color: color for this density plot
        :param show_y_label: whether to show y-axis label
        :param y_label: the text of y-axis title
        :param theme: the theme name
        :param strand_choice: the strand to draw on site plot
        :param only_customized_junction: only draw customized junctions
        :return:
        """
        obj, category = self.__init_input_file__(
            path=path,
            category=category,
            label=label,
            title=title,
            barcode=barcode,
            barcode_groups=barcode_groups,
            barcode_tag=barcode_tag,
            umi_tag=umi_tag,
            library=library,
            size_factor=size_factor,
            density_by_strand=density_by_strand,
            log_trans=log_trans
        )

        type_ = "density"
        if show_site_plot and category == "bam":
            type_ = "site-plot"
        elif show_site_plot:
            logger.warning("show_site_plot only works with bam files")

        info = PlotInfo(obj=obj, type_=type_, category=category)

        new_obj = True
        for p in self.plots:
            if p.category == info.category:
                for obj_ in p.obj:
                    if obj_.label == label and label:
                        param = self.params.pop(p)
                        p.obj.append(obj)
                        new_obj = False
                        self.params[p] = param
                        break
            if not new_obj:
                break

        if new_obj:
            self.plots.append(info)
            self.params[info] = {
                "show_junction_number": show_junction_number,
                "junction_number_font_size": junction_number_font_size,
                "color": color,
                "font_size": font_size,
                "n_y_ticks": n_y_ticks,
                "show_y_label": show_y_label,
                "y_label": y_label,
                "theme": theme,
                "strand_choice": strand_choice,
                "density_by_strand": density_by_strand,
                "only_customized_junction": only_customized_junction
            }
        return self

    def add_heatmap(self,
                    path: str,
                    group: str = "",
                    category: str = "bam",
                    size_factor=None,

                    # file loading parameters
                    label: Union[str, List[str]] = "",
                    title: str = "",
                    barcode: str = "",
                    barcode_groups: Dict[str, Set[str]] = None,
                    barcode_tag: str = "BC",
                    umi_tag: str = "UB",
                    library: str = "fru",

                    # plotting parameters
                    color="viridis",
                    font_size: int = 8,
                    show_y_label: bool = True,
                    theme: str = "ticks_blank",
                    do_scale: bool = False,
                    clustering: bool = False,
                    clustering_method: str = "ward",
                    distance_metric: str = "euclidean",
                    show_row_names: bool = False,
                    vmin=None, vmax=None,
                    log_trans: Optional[str] = None,
                    ):
        u"""
        add multiple objects for a group of heatmap
        :param path: path to input files
        :param group: the heatmap group
        :param category: file category corresponding to input file
        :param label: the label of input file
        :param size_factor: only used by atac
        :param title: the title of input file
        :param barcode: key of barcode barcode_groups
        :param barcode_groups:
        :param barcode_tag: cell barcode tag
        :param umi_tag: umi barcode tag
        :param library: fru: fr-unstrand
        :param color: color for this density plot
        :param show_y_label: whether to show y-axis label
        :param theme: the theme name
        :param font_size:
        :param do_scale: whether to scale the matrix
        :param clustering: whether reorder matrix by clustering
        :param clustering_method: same as  scipy.cluster.hierarchy.linkage
        :param distance_metric: same as scipy.spatial.distance.pdist
        :param color: used for seaborn.heatmap, see: https://matplotlib.org/3.5.1/tutorials/colors/colormaps.html
                    'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone',
                    'pink', 'spring', 'summer', 'autumn', 'winter', 'cool',
                    'Wistia', 'hot', 'afmhot', 'gist_heat', 'copper'
        :param show_row_names:
        :param vmin: Values to anchor the colormap,
                     otherwise they are inferred from the data and other keyword arguments.
        :param vmax: Values to anchor the colormap,
                     otherwise they are inferred from the data and other keyword arguments.
        :return:
        """
        obj, category = self.__init_input_file__(
            path=path,
            category=category,
            label=label,
            title=title,
            barcode=barcode,
            barcode_groups=barcode_groups,
            barcode_tag=barcode_tag,
            umi_tag=umi_tag,
            library=library,
            size_factor=size_factor,
            log_trans=log_trans
        )

        exists = False
        for p in self.plots:
            if p.group == group and p.type == "heatmap":
                p.add(obj=obj, category=category, type_="heatmap")
                exists = True
                break

        if not exists:
            info = PlotInfo(obj=obj, category=category, type_="heatmap", group=group)
            self.plots.append(info)
            self.params[info] = {
                "color": color,
                "font_size": font_size,
                "show_y_label": show_y_label,
                "theme": theme,
                "do_scale": do_scale,
                "clustering": clustering,
                "clustering_method": clustering_method,
                "distance_metric": distance_metric,
                "show_row_names": show_row_names,
                "vmin": vmin, "vmax": vmax
            }

        return self

    def add_line(self,
                 path: str,
                 group: str = "",
                 category: str = "bam",

                 # file loading parameters
                 label: Union[str, List[str]] = "",
                 title: str = "",
                 barcode: str = "",
                 barcode_groups: Dict[str, Set[str]] = None,
                 barcode_tag: str = "BC",
                 umi_tag: str = "UB",
                 library: str = "fru",

                 # plotting parameters
                 color="blue",
                 font_size: int = 8,
                 show_y_label: bool = True,
                 line_attrs: Optional[Dict] = None,
                 theme: str = "ticks_blank",
                 n_y_ticks: int = 4,
                 show_legend: bool = False,
                 legend_position: str = "upper right",
                 legend_ncol: int = 0,
                 log_trans: Optional[str] = None,
                 ):
        u"""
        add multiple objects for a group of heatmap
        :param path: path to input files
        :param group: the heatmap group
        :param category: file category corresponding to input file
        :param label: the label of input file
        :param title: the title of input file
        :param barcode: key of barcode barcode_groups
        :param barcode_groups:
        :param barcode_tag: cell barcode tag
        :param umi_tag: umi barcode tag
        :param library: fru: fr-unstrand
        :param n_y_ticks: number of y ticks
        :param color: color for this density plot
        :param show_y_label: whether to show y-axis label
        :param theme: the theme name
        :param font_size:
        :param line_attrs: the additional attributes to control the line, usd by matpltolib.axes.Axes.plot
        :param show_legend: whether to show legend
        :param legend_position:
        :param legend_ncol:
        :return:
        """
        obj, category = self.__init_input_file__(
            path=path,
            category=category,
            label=label,
            title=title,
            barcode=barcode,
            barcode_groups=barcode_groups,
            barcode_tag=barcode_tag,
            umi_tag=umi_tag,
            library=library,
            log_trans=log_trans
        )

        if not line_attrs:
            line_attrs = {}
        line_attrs["color"] = color

        exists = False
        for p in self.plots:
            if p.group == group and p.type == "line":
                p.add(obj=obj, category=category, type_="line")
                self.params[p]["line_attrs"][obj.label] = line_attrs
                exists = True
                break

        if not exists:
            info = PlotInfo(obj=obj, category=category, type_="line", group=group)
            self.plots.append(info)
            self.params[info] = {
                "line_attrs": {obj.label: line_attrs},
                "show_legend": show_legend,
                "font_size": font_size,
                "n_y_ticks": n_y_ticks,
                "show_y_label": show_y_label,
                "theme": theme,
                "legend_position": legend_position,
                "legend_ncol": legend_ncol
            }

        return self

    def add_hic(
            self,
            path: str,
            category: str = "hic",
            label: str = "",
            color: str = "RdYlBu_r",
            trans: Optional[str] = None,
            show_legend: bool = True,
            depth: int = 30000,
            font_size: int = 8,
            n_y_ticks: int = 4,
            show_y_label: bool = True,
            theme: str = "ticks"
    ):
        obj, category = self.__init_input_file__(
            path=path,
            category=category,
            label=label,
            depth=depth,
            trans=trans
        )

        info = PlotInfo(obj=obj, category=category, type_="hic")
        self.plots.append(info)
        self.params[info] = {
            "show_legend": show_legend,
            "color": color,
            "y_label": label,
            "font_size": font_size,
            "n_y_ticks": n_y_ticks,
            "show_y_label": show_y_label,
            "theme": theme
        }

        return self

    def add_igv(
            self,
            path: str,
            category: str = "igv",
            label: str = "",
            exon_focus: Optional[str] = None,

            # file loading parameters
            library: str = "fru",
            features: Optional[dict] = None,
            deletion_ignore: Optional[int] = True,
            del_ratio_ignore: float = .5,

            # plotting parameters
            exon_color: Optional[str] = None,
            intron_color: Optional[str] = None,
            feature_color: Optional[str] = None,
            exon_width: float = .3,
            font_size: int = 8,
            n_y_ticks: int = 1,
            show_y_label: bool = True,
            theme: str = "ticks_blank"
    ):
        u"""
        Add igv-like plot into track
        :param path: path to input files
        :param category: file category for the input file
        :param library: fru: fr-unstrand
        :param features:
        :param exon_focus:
        :param deletion_ignore:
        :param del_ratio_ignore:
        :param label:
        :param exon_color:
        :param intron_color:
        :param feature_color:
        :param exon_width:
        :param font_size:
        :param n_y_ticks:
        :param show_y_label:
        :param theme:
        :return:
        """
        obj, category = self.__init_input_file__(
            path=path,
            category=category,
            library=library,
            features=features,
            deletion_ignore=deletion_ignore,
            del_ratio_ignore=del_ratio_ignore,
            exon_focus=exon_focus
        )

        info = PlotInfo(obj=obj, category=category, type_="igv")
        self.plots.append(info)
        self.params[info] = {
            "y_label": label,
            "exon_color": exon_color,
            "intron_color": intron_color,
            "feature_color": feature_color,
            "exon_width": exon_width,
            "font_size": font_size,
            "n_y_ticks": n_y_ticks,
            "show_y_label": show_y_label,
            "theme": theme
        }

        return self

    def add_motif(self,
                  path: str,
                  category: str = "motif",
                  motif_region: GenomicLoci = None,
                  width: float = 0.8,
                  theme: str = "blank",
                  **kwargs):
        obj = Motif.create(path, self.region)

        info = PlotInfo(obj=obj, category=category, type_="motif")

        self.plots.append(info)
        self.params[info] = {"width": width, "theme": theme, "region": motif_region if motif_region else self.region}
        return self

    def add_manual(self,
                   data: np.array,
                   image_type: str = "line",
                   label: str = "",
                   group: str = "",
                   color: str = "blue",
                   font_size: int = 8,
                   n_y_ticks: int = 1,
                   show_y_label: bool = True,
                   theme: str = "ticks_blank",
                   **kwargs
                   ):
        obj = File("")
        obj.label = label
        obj.data = ReadDepth(data)
        info = PlotInfo(obj=obj, category="manual", type_=image_type, group=group)

        exists = False
        if group:
            for p in self.plots:
                if p.group == group and p.type == image_type:
                    p.add(obj=obj, category="manual", type_=image_type)
                    self.params[p]["line_attrs"][obj.label] = {"color": color}
                    exists = True
                    break

        if not exists:
            self.plots.append(info)
            self.params[info] = {
                "font_size": font_size,
                "n_y_ticks": n_y_ticks,
                "show_y_label": show_y_label,
                "theme": theme,
                "color": color,
                "line_attrs": {obj.label: {"color": color}}
            }
            self.params[info].update(kwargs)

        return self

    def merge_by_cell(self):

        plots = {}
        for p in self.plots:
            assert isinstance(p, PlotInfo)

            if p.category in ["density", "line"]:
                label = p.obj[0].label
                if label in plots.keys():
                    plots[label] += p
                else:
                    plots[label] = p
            else:
                plots[p.obj[0].label] = p

        self.plots = list(plots.values())

        return self

    def __len__(self) -> int:
        return self.end - self.start + 1

    def copy(self):
        return deepcopy(self)

    def plot(self,
             output: Optional[str] = None,
             reference_scale: Union[int, float] = .25,
             stroke_scale: Union[int, float] = .25,
             dpi: int = 300,
             width: Union[int, float] = 0,
             height: Union[int, float] = 0,
             raster: bool = False,
             return_image: Optional[str] = None,
             sc_height_ratio: Optional[Dict[str, float]] = None,
             distance_between_label_axis: float = .3,
             n_jobs: int = 1,
             *args, **kwargs):
        u"""
        draw image
        :param output: if output is empty then show this image by plt.showfig
        :param reference_scale: to adjust the size of reference plot
        :param stroke_scale: to adjust the size of stroke plot
        :param dpi: the dpi of saved plot
        :param width: the width of figure, if width == 0, the let matplotlib decide the size of image
        :param height: the height of figure, if height == 0, the let matplotlib decide the size of image
        :param raster: plot rasterizer site plot
        :param sc_height_ratio: adjust the relative height of single cell plots
        :param distance_between_label_axis: distance between y-axis label and y-axis ticks
        :param n_jobs: load data in how many processes
        :param return_image: used for interactive ui
        """
        if sc_height_ratio is None:
            sc_height_ratio = {"density": .2, "heatmap": .2}
        assert self.region is not None, f"please set the plotting region first."

        plots_n_rows = 1
        plots_n_cols = 1

        height_ratio = []
        if self.reference is not None:
            logger.info("load reference")
            self.reference.load(self.region, *args, **kwargs)
            plots_n_rows += self.reference.len(scale=reference_scale)

        if self.stroke:
            plots_n_rows += int(max(len(self.stroke) * stroke_scale, 1))

        if self.link:
            plots_n_rows += len(self.link)

        if self.sequence is not None:
            logger.info("load sequence")
            self.sequence.load(self.region)

        logger.info(f"load data of {len(self.plots)} plots")

        cmds = []
        for p in self.plots:
            assert isinstance(p, PlotInfo), f"unrecognized data type: {type(p)}"
            # only enable the multiprocessing while n_jobs > 1
            if self.__n_objs__ / len(self.plots) >= n_jobs > 1:
                p.load(self.region, n_jobs, junctions=self.junctions.get(p.obj[0].label, {}), *args, **kwargs)
            elif n_jobs > 1:
                temp = copy.deepcopy(kwargs)
                temp["region"] = self.region if p.type != "motif" else self.params[p]["region"]
                temp["junctions"] = self.junctions.get(p.obj[0].label, {})
                cmds.append([p, args, temp])

        if len(cmds) > 0:
            with Pool(max(1, min(n_jobs, len(self.plots)))) as p:
                self.plots = p.map(__load__, cmds)

        # count the plots size
        for p in self.plots:
            if n_jobs <= 1:
                p.load(self.region, junctions=self.junctions.get(p.obj[0].label, {}), *args, **kwargs)

            for obj in p.obj:
                obj.transform()

            plots_n_rows += p.len(reference_scale)
            if p.type in ["heatmap", "hic"]:
                plots_n_cols = 2

            if p.is_single_cell:
                height_ratio.append(sc_height_ratio.get(p.type, 1))
            else:
                height_ratio.append(1)

        logger.debug(f"plots n_rows={plots_n_rows}; n_cols = {plots_n_cols}")
        logger.info("init graph_coords")
        exon_scale = kwargs.get("exon_scale", 1)
        intron_scale = kwargs.get("intron_scale", .5)
        if plots_n_cols > 1 and intron_scale != 1:
            logger.warning(f"heatmap require intron_scale = 1")
            intron_scale = 1

        self.graph_coords = init_graph_coords(
            self.region, self.exons,
            exon_scale=exon_scale,
            intron_scale=intron_scale
        )

        height_ratio += [1 for _ in range(plots_n_rows - len(height_ratio))]

        if width and height:
            fig = plt.figure(figsize=[width, height * sum(height_ratio)], dpi=dpi)
        else:
            fig = plt.figure(dpi=dpi)

        if plots_n_cols > 1:
            gs = gridspec.GridSpec(plots_n_rows, plots_n_cols, height_ratios=height_ratio,
                                   width_ratios=(.99, .01), wspace=0.01, hspace=.15)
        else:
            gs = gridspec.GridSpec(plots_n_rows, plots_n_cols, height_ratios=height_ratio,
                                   wspace=.7, hspace=.15)

        max_used_y_val = None
        if kwargs.get("same_y"):
            for p in self.plots:
                if p.type in ["density", "site-plot", "line"]:
                    for obj in p.obj:
                        y = max(obj.data.wiggle)
                        max_used_y_val = y if max_used_y_val is None else max(y, max_used_y_val)

        curr_idx = 0
        for p in self.plots:
            if p.type == "igv":
                ax_var = plt.subplot(gs[curr_idx: curr_idx + p.len(reference_scale), 0])
            else:
                ax_var = plt.subplot(gs[curr_idx, 0])
            if p.type == "density":
                plot_density(
                    ax=ax_var,
                    obj=p.obj[0],
                    graph_coords=self.graph_coords,
                    max_used_y_val=max_used_y_val,
                    distance_between_label_axis=distance_between_label_axis,
                    raster=raster,
                    **self.params[p]
                )
            elif p.type == "hic":
                plot_hic(
                    ax=ax_var,
                    cbar_ax=plt.subplot(gs[curr_idx, 1]),
                    obj=p.obj,
                    distance_between_label_axis=distance_between_label_axis,
                    raster=raster,
                    **self.params[p]
                )
            elif p.type == "site-plot":
                plot_density(
                    ax=ax_var,
                    obj=p.obj[0],
                    graph_coords=self.graph_coords,
                    max_used_y_val=max_used_y_val,
                    distance_between_label_axis=distance_between_label_axis,
                    raster=raster,
                    **self.params[p]
                )

                site_ax = plt.subplot(gs[curr_idx + 1, 0])

                plot_site_plot(
                    site_ax, p.obj[0],
                    graph_coords=self.graph_coords,
                    raster=raster,
                    distance_between_label_axis=distance_between_label_axis,
                    **self.params[p]
                )
                curr_idx += 1
            elif p.type == "heatmap":
                plot_heatmap(
                    ax=ax_var,
                    cbar_ax=plt.subplot(gs[curr_idx, 1]),
                    data=p.data,
                    y_label=p.group,
                    graph_coords=self.graph_coords,
                    raster=raster,
                    distance_between_label_axis=distance_between_label_axis,
                    **self.params[p]
                )
            elif p.type == "line":
                plot_line(
                    ax=ax_var,
                    data=p.data,
                    y_label=p.group,
                    max_used_y_val=max_used_y_val,
                    graph_coords=self.graph_coords,
                    distance_between_label_axis=distance_between_label_axis,
                    **self.params[p]
                )
            elif p.type == "igv":
                plot_igv_like(
                    ax=ax_var,
                    obj=p.data,
                    graph_coords=self.graph_coords,
                    raster=raster,
                    distance_between_label_axis=distance_between_label_axis,
                    **self.params[p]
                )
            elif p.type == "motif":
                plot_motif(ax=ax_var, obj=p.obj[0], graph_coords=self.graph_coords, **self.params[p])
            else:
                raise ValueError(f"unknown plot type {p.type}")

            # adjust indicator lines and focus background
            set_indicator_lines(ax=ax_var, sites=self.sites, graph_coords=self.graph_coords)
            set_focus(ax=ax_var, focus=self.focus, graph_coords=self.graph_coords)

            if p.type != "igv":
                curr_idx += 1
            else:
                curr_idx += p.len(reference_scale)

        if self.link:
            for link in self.link:
                plot_links(ax=plt.subplot(gs[curr_idx:(curr_idx + 1), 0]),
                           data=link, graph_coords=self.graph_coords)
                curr_idx += 1

        # draw x label
        set_x_ticks(
            ax=plt.subplot(gs[curr_idx, 0]), region=self.region,
            graph_coords=self.graph_coords,
            sequence=self.sequence.data if self.sequence else None,
            *args, **kwargs
        )
        curr_idx += 1

        if self.reference is not None:
            ax_var = plt.subplot(gs[curr_idx:curr_idx + self.reference.len(scale=reference_scale), 0])
            plot_reference(ax=ax_var, obj=self.reference,
                           graph_coords=self.graph_coords,
                           plot_domain=self.reference.add_domain,
                           distance_between_label_axis=distance_between_label_axis,
                           **self.params["reference"])

            # adjust indicator lines and focus background
            set_indicator_lines(ax=ax_var, sites=self.sites, graph_coords=self.graph_coords)
            set_focus(ax=ax_var, focus=self.focus, graph_coords=self.graph_coords)
            curr_idx += self.reference.len(scale=reference_scale)

        if self.stroke:
            ax_var = plt.subplot(gs[curr_idx:plots_n_rows, 0])
            plot_stroke(ax=ax_var, data=self.stroke, graph_coords=self.graph_coords, *args, **kwargs)

        if output:
            fig.savefig(output, transparent=True, bbox_inches='tight')
        elif return_image:
            output = io.BytesIO()
            if return_image == "png":
                FigureCanvasAgg(fig).print_png(output)
            elif return_image == "pdf":
                FigureCanvasPdf(fig).print_pdf(output)
            return output
        else:
            plt.show()

        plt.close()
        return self


if __name__ == '__main__':
    pass



import os
from setuptools import setup, find_packages

__version__ = "0.1.2"
__author__ = "ygidtu & Ran Zhou"
__email__ = "ygidtu@gmail.com"


def locate_packages():
    __dir__ = os.path.dirname(os.path.abspath(__file__))
    pipfile = os.path.join(__dir__, "Pipfile")
    packages = []

    kept = False
    with open(pipfile) as r:
        for line in r:
            line = line.strip()

            if line == "[packages]":
                kept = True
                continue
            elif line.startswith("[") and line.endswith("]"):
                kept = False

            if kept:
                if line:
                    name, version = line.split(" = ")
                    version = version.strip('"')
                    if version == "*":
                        packages.append(f"{name}")
                    else:
                        packages.append(f"{name}, {version}")
    return packages


def load_description():
    __dir__ = os.path.dirname(os.path.abspath(__file__))
    readme = os.path.join(__dir__, "README.md")
    if os.path.exists(readme):
        with open(readme) as r:
            return r.read()
    return "README"


setup(
    name='sashimi.py',
    long_description=load_description(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    url="https://github.com/ygidtu/sashimi.py",
    entry_points={
        'console_scripts':
            [
                'sashimipy = sashimi.cli:main'
            ]
    },
    python_requires='>=3.6',
    data_files=[(".", ['README.md', 'Pipfile', 'Pipfile.lock'])],
    install_requires=locate_packages(),
    version=__version__
)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
Created at 2022.03.22 by ygidtu@gmail.com

This is used to create heatmap images for multiple signal files contains in bigWig file
"""
import os

import numpy as np

from sashimi.base.GenomicLoci import GenomicLoci
from sashimi.base.ReadDepth import ReadDepth
from sashimi.base.Readder import Reader
from sashimi.file.File import File


class Bedgraph(File):

    def __init__(self, path: str, label: str = "", title: str = ""):
        u"""
        :param path: the path to bam file
        :param label: the left axis label
        """
        super().__init__(path)
        self.label = label
        self.title = title

    @classmethod
    def create(cls, path: str, label: str = "", title: str = ""):
        assert os.path.exists(path), f"{path} is not exists."
        if not label:
            label = os.path.basename(path)
        return cls(path=path, label=label, title=title)

    def load(self, region: GenomicLoci, **kwargs):
        self.region = region

        depth = np.zeros(len(self.region))

        for record in Reader.read_depth(self.path, self.region):
            if len(record) >= 4:
                _, start, end, cov = record[:4]
                start, end, cov = int(start), int(end), float(cov)
                depth[(start - self.region.start):(end - self.region.start)] += cov
        self.data = ReadDepth(depth)


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
This file contains the configuration of different themes
"""
from matplotlib import axes


class Theme(object):
    @classmethod
    def blank(cls, ax: axes.Axes):
        u"""

        :param ax:
        :return:
        """
        # ax.spines['left'].set_position('zero')
        # ax.spines['bottom'].set_position('zero')
        # ax.spines['top'].set_visible(False)
        # ax.spines['right'].set_visible(False)
        ax.set_axis_off()
        ax.tick_params(bottom=False, top=False, left=False, right=False)

    @classmethod
    def ticks(cls, ax: axes.Axes):
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.tick_params(bottom=True, top=False, left=True, right=False)

    @classmethod
    def ticks_blank(cls, ax: axes.Axes):
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.set_xticklabels([])
        ax.tick_params(bottom=False, top=False, left=True, right=False)

    @classmethod
    def set_theme(cls, ax: axes.Axes, name: str = "blank"):
        for i, func in cls.__dict__.items():
            if isinstance(func, classmethod) and i != "get" and i == name:
                """
                This usage see: https://blog.peterlamut.com/tag/classmethod/
                """
                return func.__get__(None, cls)(ax)

        raise ValueError(f"{name} is not a valid theme")


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
This file contains the object to handle bam file related issues.

changelog:
    1. add library parameter for determining of read strand at 2022.4.28.

"""
import os
from typing import Dict, Optional, Set

import numpy as np
from loguru import logger

from sashimi.base.GenomicLoci import GenomicLoci
from sashimi.base.ReadDepth import ReadDepth
from sashimi.base.Readder import Reader
from sashimi.file.File import SingleCell


class ATAC(SingleCell):
    def __init__(self,
                 path: str, barcode_groups: Dict[str, Set[str]], barcode: str, size_factor,
                 label: str = "", title: str = ""):
        u"""
        init this object
        :param label: the left axis label
        :param title: the default title to show in the upper-right of density plot
        :param barcode_groups:
        :param barcode: key of barcode_groups
        :param size_factor
        """
        super().__init__(path, barcode_groups[barcode])
        self.title = title
        self.label = label if label else os.path.basename(path).replace(".bam", "")

        self.size_factor = size_factor
        self.barcode_groups = barcode_groups
        self.barcode = barcode

    @classmethod
    def index(cls, path: str, barcodes: Optional[Dict[str, Set[str]]] = None):
        reverse_barcode_groups = {}  # dict, key => barcode, value => group
        for key, vals in barcodes.items():
            for val in vals:
                reverse_barcode_groups[val.strip()] = key.strip()
        size_factors, sizes = {}, {}
        for values in Reader.read_depth(path=path):
            values = values.split()
            count = int(values[-1])
            barcode = values[-2]
            if barcode not in reverse_barcode_groups.keys():
                continue

            key = reverse_barcode_groups[barcode]

            sizes[key] = sizes.get(key, 0) + 1
            size_factors[key] = size_factors.get(key, 0) + count

        for key, val in size_factors.items():
            size_factors[key] = size_factors[key] * sizes[key]

        del sizes
        median_size_factor = np.median(np.array(list(size_factors.values())))
        return {x: y / median_size_factor for x, y in size_factors.items()}

    @classmethod
    def create(cls,
               path: str,
               label: str = "",
               title: str = "",
               barcode_groups: Optional[Dict[str, Set[str]]] = None,
               barcode: Optional[str] = None,
               size_factors=None):
        u"""
        :param path: the path to bam file
        :param label: the left axis label
        :param title: the default title to show in the upper-right of density plot
        :param barcode_groups:
        :param barcode: key of barcode_groups:
        :param size_factors
        :return:
        """

        return cls(
            path=path,
            label=label,
            title=title,
            barcode=barcode,
            barcode_groups=barcode_groups,
            size_factor=size_factors
        )

    def __hash__(self):
        return hash(self.label)

    def __str__(self) -> str:

        temp = []

        for x in [self.title, self.label, self.path]:
            if x is None or x == "":
                x = "None"
            temp.append(str(x))

        return "\t".join(temp)

    def load(self,
             region: GenomicLoci,
             threshold: int = 0,
             required_strand: Optional[str] = None,
             log_trans: Optional[str] = None,
             **kwargs):
        """
            determine_depth determines the coverage at each base between start_coord and end_coord, inclusive.

            bam_file_path is the path to the bam file used to \
            determine the depth and junctions on chromosome between start_coord and end_coord

        return values:
            depth_vector,
            which is a Numpy array which contains the coverage at each base position between start_coord and end_coord

            spanned_junctions, which is a dictionary containing the junctions supported by reads.
            The keys in spanned_junctions are the
                names of the junctions, with the format chromosome:lowerBasePosition-higherBasePosition
        :param region: GenomicLoci object including the region for calculating coverage
        :param threshold: minimums counts of the given splice junction for visualization
        :param reads1: None -> all reads, True -> only R1 kept; False -> only R2 kept
        :param required_strand: None -> all reads, else reads on specific strand
        :param log_trans: should one of {"10": np.log10, "2": np.log2}
        """
        self.region = region
        self.log_trans = log_trans

        depth_vector = np.zeros(len(region), dtype=int)
        try:
            for _, start, end, barcode, count in Reader.read_depth(path=self.path, region=region):
                # filter reads by 10x barcodes
                start, end, count = int(start), int(end), int(count)
                if not self.empty_barcode():
                    if not self.has_barcode(barcode):
                        continue
                depth_vector[(start - region.start)] += count
                depth_vector[(end - region.start)] += count
        except IOError as err:
            logger.error('There is no .bam file at {0}'.format(self.path))
            logger.error(err)
        except ValueError as err:
            logger.error(self.path)
            logger.error(err)

        self.data = ReadDepth(depth_vector)
        self.data.normalize(self.size_factor[self.barcode])
        return self


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
Created by ygidtu@gmail.com at 2020.05.07

This scripts contains the class handle the reference file
"""
import glob
import gzip
import os
import re
from collections import namedtuple, defaultdict
from typing import List, Union, Optional

import filetype
import pysam
from loguru import logger

from sashimi.base.GenomicLoci import GenomicLoci
from sashimi.base.Protein import CdsProtein
from sashimi.base.Readder import Reader
from sashimi.base.Transcript import Transcript
from sashimi.file.File import File


class Reference(File):
    u"""
    The reference file, support gtf and gff format
    """

    def __init__(self, path: str, category: str = "gtf",
                 add_domain: bool = False, add_local_domain: Optional[str] = False,
                 proxy: Optional[str] = None, timeout: int = 10):
        u"""
        init func
        :param path: path to input file
        """
        categories = ["gtf", "bam", "bed"]
        assert category in categories, f"category should be one of {categories}, instead of {category}"

        super().__init__(path=self.index_gtf(path) if category == "gtf" else path)
        self.category = category
        self.data = []
        self.add_domain = add_domain
        self.domain = None
        self.interval_file = {}

        self.add_local_domain = add_local_domain
        self.local_domain = None

        self.proxy = proxy
        self.timeout = timeout

        if proxy:
            logger.info(f"Using proxy: {proxy}")

    def __add__(self, other):
        assert isinstance(other, Reference), "only Reference and Reference could be added"
        new_ref = Reference(self.path, category=self.category)
        new_ref.data += self.data
        new_ref.data += other.data
        new_ref.data = sorted(new_ref.data)

        new_ref.interval_file.update(other.interval_file)

        if self.domain:
            new_ref.__add_domain__()
        if self.add_local_domain:
            new_ref.__load_local_domain__()
        return new_ref

    def len(self, scale: Union[int, float] = .25) -> int:
        u"""
        the length of reference to draw in final plots, default using the quarter of number of transcripts
        """
        size = len(self.data)

        if self.domain:
            size += len(self.domain)

        return int(max(size * float(scale), 1))

    @property
    def exons(self) -> List[List[int]]:
        res = []
        for transcript in self.data:
            for exon in transcript.exons:
                res.append([exon.start, exon.end])
        return sorted(res, key=lambda x: [x[0], x[1]])

    @classmethod
    def create(cls, path: str, add_domain: bool = False, add_local_domain: Optional[str] = False,
               category: str = "gtf"):
        u"""
        create reference file object
        :param path: path to input file
        :param add_domain: whether plot domain
        :param add_local_domain: fetch domain information from local file, which download from ucsc
        :param category: the type of reference file, include gtf and bam: customized reads as references
        :return: Reference obj
        """
        assert os.path.exists(path), f"{path} not exists"

        if add_local_domain:
            assert os.path.isdir(add_local_domain), f"{add_local_domain} not exists"

        return cls(
            path=path,
            add_domain=add_domain,
            add_local_domain=add_local_domain,
            category=category)

    def __load_local_domain__(self, region: GenomicLoci):

        if self.local_domain:
            u"""
            Because there is no gene or transcript id, we don't merge multiple domain here.
            """
            pass

        Domain_region = namedtuple(
            "DomainGenomicRegion",
            ["category", "type", "description", "unique_id", "start", "end"]
        )
        protein_info = defaultdict(list)

        if self.add_local_domain:
            for bb_file in glob.glob(pathname=f"{self.add_local_domain}/*.bb"):
                domain_res = defaultdict(list)
                base_name = os.path.basename(bb_file).replace(".bb", "")
                for record in Reader.read_bigbed(bb_file, region):
                    record = record[2].split("\t")
                    current_id = record[0]
                    current_start = int(record[3])
                    block_sizes = [int(x) for x in record[7].split(",") if x]
                    block_starts = [int(x) for x in record[8].split(",") if x]
                    current_desc = record[17]

                    current_domain_res = []
                    for index in range(len(block_starts)):
                        current_domain_res.append(
                            Domain_region._make(
                                [current_id,
                                 current_id,
                                 current_desc,
                                 current_id,
                                 current_start + 1 + block_starts[index],
                                 current_start + 1 + block_starts[index] + block_sizes[index] - 1
                                 ]
                            )
                        )
                    domain_res[current_id].extend([current_domain_res])

                for domain_unique_id, domain_list in domain_res.items():
                    start_site = min([
                        min(map(lambda x: x.start, i)) for i in domain_list
                    ])

                    end_site = max([
                        max(map(lambda x: x.end, i)) for i in domain_list
                    ])

                    protein_info[base_name].append(
                        Transcript(
                            chromosome=region.chromosome,
                            start=start_site,
                            end=end_site,
                            strand="*",
                            exons=domain_list,
                            gene=domain_list[0][0].unique_id,
                            domain_type=domain_list[0][0].type,
                            domain_description=domain_list[0][0].description,
                            domain_category=domain_list[0][0].category,
                            category="protein"
                        )
                    )

            self.local_domain = protein_info

    def __add_domain__(self):
        gene_id = set(map(lambda x: x.gene_id, self.data))
        transcript_id = set(map(lambda x: x.transcript_id, self.data))
        chromosome_id = self.data[0].chromosome

        # if domain is not None, then add it.
        if self.domain:
            gene_id = gene_id.difference(self.domain.gene_id)
            transcript_id = transcript_id.difference(self.domain.transcript_id)

        if len(transcript_id) != 0 and len(gene_id) != 0:
            with pysam.Tabixfile(self.path) as gtf_tabix:
                domain_info = CdsProtein.__re_iter_gtf__(
                    gtf_tabix=gtf_tabix,
                    chromosome=chromosome_id,
                    transcript_id=transcript_id,
                    gene_id=chromosome_id
                )
            # add pep information to cdsProtein.
            if self.domain:
                self.domain.add(domain_info)
            else:
                self.domain = domain_info

    def add_interval(self, interval: str, label: str):
        u"""
        Add another annotation in to data, and each annotation has a track
        :param interval: a bed file
        :param label
        :return:
        """
        assert os.path.exists(interval), f"{interval} not exists."
        self.interval_file[interval] = label
        return self

    @staticmethod
    def is_gtf(infile):
        u"""
        check if input file is gtf
        :param infile: path to input file
        :return:
        """
        if infile is None:
            return False

        is_gtf = 0
        try:
            if filetype.guess_mime(infile) == "application/gzip":
                is_gtf += 10
                r = gzip.open(infile, "rt")
            else:
                r = open(infile)

            for line in r:
                if line.startswith("#"):
                    continue

                lines = re.split(r"\s+", line)

                if len(lines) < 8:
                    break

                if re.search(r"([\w-]+ \"[\w.\s\-%,:]+\";? ?)+", " ".join(lines[8:])):
                    is_gtf += 1

                break

            r.close()
            return is_gtf
        except TypeError as err:
            logger.error("failed to open %s", infile)
            exit(err)

    @classmethod
    def sort_gtf(cls, input_gtf: str, output_gtf: str):

        if os.path.exists(output_gtf) and os.path.getsize(output_gtf) > 10:
            logger.info(f"{output_gtf} already exists, skip sorting")
            return

        data = []
        logger.info("Sorting %s" % input_gtf)

        try:
            w = open(output_gtf.replace(".gz", ""), "w+")
            r = gzip.open(input_gtf, "rt") if input_gtf.endswith(".gz") else open(input_gtf)
            for line in r:
                if line.startswith("#"):
                    w.write(line)
                    continue

                lines = line.split()

                if len(lines) < 1:
                    continue

                data.append(
                    GenomicLoci(
                        chromosome=lines[0], start=lines[3], end=lines[4],
                        strand=lines[6], gtf_line=line
                    )
                )
            r.close()

            for i in sorted(data, key=lambda x: [x.chromosome, x.start]):
                w.write(i.gtf_line)
            w.close()

            pysam.tabix_compress(output_gtf.replace(".gz", ""), output_gtf)
            os.remove(output_gtf.replace(".gz", ""))
        except IOError as err:
            logger.error(f"could not sort gtf, because {err}")
            exit(err)

    @classmethod
    def index_gtf(cls, input_gtf):
        u"""
        Check the tabix index of input gtf file

        Extract only exon tags and keep it clean
        :param input_gtf: path to input gtf file
        :return path to compressed and indexed bgzipped gtf file
        """
        gtf = cls.is_gtf(input_gtf)
        assert gtf % 10 == 1, f"{input_gtf} seems not be gtf format"

        if gtf // 10 > 0:
            output_gtf = input_gtf
        else:
            output_gtf = input_gtf + ".gz"

        if not os.path.exists(output_gtf + ".tbi"):
            logger.info(f"Create index for {input_gtf}")
            try:
                pysam.tabix_index(
                    output_gtf,
                    preset="gff",
                    force=True,
                    keep_original=True
                )
            except OSError as err:
                logger.error(err)
                logger.error("Guess gtf needs to be sorted")

                sorted_gtf = re.sub(r"\.gtf(.gz)?$", "", input_gtf) + ".sorted.gtf.gz"
                if os.path.exists(sorted_gtf) and os.path.exists(sorted_gtf + ".tbi"):
                    return sorted_gtf

                cls.sort_gtf(input_gtf, sorted_gtf)
                pysam.tabix_index(
                    sorted_gtf, preset="gff",
                    force=True, keep_original=True
                )
                return sorted_gtf
        elif os.path.getctime(output_gtf) < os.path.getctime(output_gtf):
            logger.info("the tbi index is older than the gtf file")

        return output_gtf

    def __load_gtf__(self, region: GenomicLoci) -> List[Transcript]:

        u"""
        Load transcripts inside of region from gtf file
        :param region: target region
        :return: list of Transcript
        """
        transcripts = {}
        exons = {}

        for rec in Reader.read_gtf(self.path, region):
            start = max(rec.start, region.start)
            end = min(rec.end, region.end)

            if end + 1 <= region.start:
                continue
            if start + 1 >= region.end:
                break

            if re.search(r"(rna|transcript|cds)", rec.feature, re.I):
                if rec.transcript_id not in transcripts.keys():
                    transcripts[rec.transcript_id] = Transcript(
                        chromosome=rec.contig,
                        start=start,
                        end=end,
                        strand=rec.strand,
                        transcript_id=rec.transcript_id,
                        gene_id=rec.gene_id,
                        gene=rec.gene_name if "gene_name" in rec.attributes else "",
                        transcript=rec.transcript_name if "transcript_name" in rec.attributes else "",
                        exons=[]
                    )
            elif re.search(r"(exon)", rec.feature, re.I):
                if rec.transcript_id not in exons.keys():
                    exons[rec.transcript_id] = []

                exons[rec.transcript_id].append(
                    GenomicLoci(
                        chromosome=rec.contig,
                        start=start,
                        end=end,
                        strand=rec.strand
                    )
                )

        for key, trans in transcripts.items():
            if key in exons.keys():
                trans.exons += exons[key]

        return sorted(transcripts.values())

    def __load_bam__(self, region: GenomicLoci, threshold_of_reads: int = 0) -> List[Transcript]:
        u"""
        Load transcripts inside of region from bam file
        :param region: target region
        :param threshold_of_reads: only kept reads with minimum frequency
        :return: list of Transcript
        """
        transcripts = {}
        try:
            for read, strand in Reader.read_bam(self.path, region):
                start = read.reference_start

                exons_in_read = []
                for cigar, length in read.cigartuples:
                    cur_start = start + 1
                    cur_end = start + length + 1

                    if cigar == 0:  # M
                        if cur_start < region.end < cur_end:
                            exons_in_read.append(GenomicLoci(
                                chromosome=read.reference_name,
                                start=cur_start if cur_start > region.start else region.start,
                                end=cur_end if cur_end <= region.end else region.end,
                                strand="+",
                            ))

                    elif cigar not in (1, 2, 4, 5):  # I, D, S, H
                        start += length

                t = Transcript(
                    chromosome=read.reference_name,
                    start=read.reference_start + 1 if read.reference_start + 1 > region.start else region.start,
                    end=read.reference_end + 1 if read.reference_end + 1 < region.end else region.end,
                    strand=strand,
                    exons=exons_in_read
                )
                transcripts[t] = transcripts.get(t, 0) + 1
        except IOError as err:
            logger.error('There is no .bam file at {0}'.format(self.path))
            logger.error(err)
        except ValueError as err:
            logger.error(self.path)
            logger.error(err)

        return sorted([x for x, y in transcripts.items() if y > threshold_of_reads])

    def __load_bed__(self, region: GenomicLoci) -> List[Transcript]:
        transcripts = []
        try:
            for rec in Reader.read_gtf(self.path, region=region, bed=True):
                exon_bound = []
                current_start = int(rec[1])
                current_end = int(rec[2])
                if len(rec) > 3:
                    current_id = rec[3]
                else:
                    current_id = "NoID"

                if len(rec) != 12:
                    exon_bound.append(
                        GenomicLoci(
                            chromosome=self.region.chromosome,
                            start=current_start + 1,
                            end=current_end,
                            strand=self.region.strand,
                            name="exon"
                        )
                    )
                else:

                    block_sizes = [int(x) for x in rec[10].split(",") if x]
                    block_starts = [int(x) for x in rec[11].split(",") if x]

                    for i in range(len(block_starts)):
                        exon_bound.append(
                            GenomicLoci(
                                chromosome=self.region.chromosome,
                                start=current_start + 1 + block_starts[i],
                                end=current_start + 1 + block_starts[i] + block_sizes[i] - 1,
                                strand=self.region.strand,
                                name="exon"
                            )
                        )

                read = Transcript(
                    chromosome=self.region.chromosome,
                    start=min(map(lambda x: x.start, exon_bound)),
                    end=max(map(lambda x: x.end, exon_bound)),
                    strand=self.region.strand,
                    transcript_id=current_id,
                    exons=exon_bound,
                )
                if read.start < self.region.start or read.end > self.region.end:
                    continue

                transcripts.append(read)

        except IOError as err:
            logger.error('There is no .bed file at {0}'.format(self.path))
            logger.error(err)
        except ValueError as err:
            logger.error(self.path)
            logger.error(err)
        return transcripts

    def load(self, region: GenomicLoci, threshold_of_reads: int = 0, **kwargs):
        u"""
        Load transcripts inside of region
        :param region: target region
        :param threshold_of_reads: used for bam file, only kept reads with minimum frequency
        :return:
        """
        assert isinstance(region, GenomicLoci), "region should be a GenomicLoci object"
        self.region = region
        if self.category == "gtf":
            self.data = self.__load_gtf__(region)
            if self.add_local_domain:
                self.__load_local_domain__(region)

            if self.add_domain:
                self.__add_domain__()
        elif self.category == "bam":
            self.data = self.__load_bam__(region, threshold_of_reads)
        elif self.category == "bed":
            self.data = self.__load_bed__(region)

        rec, start, end, strand = None, None, None, None
        for interval_file, interval_label in self.interval_file.items():
            try:
                if not os.path.exists(interval_file + ".tbi"):
                    interval_file = pysam.tabix_index(
                        interval_file,
                        preset="bed",
                        force=True,
                        keep_original=True
                    )

                interval_target = []
                for rec in Reader.read_gtf(interval_file, region=self.region, bed=True):
                    start = max(rec.start, self.region.start)
                    end = min(rec.end, self.region.end)

                    if end + 1 <= self.region.start:
                        continue
                    if start + 1 >= self.region.end:
                        break

                    try:
                        strand = rec.strand
                        rec_name = rec.name
                    except KeyError:
                        strand = "*"
                        rec_name = ""

                    interval_target.append(
                        GenomicLoci(
                            chromosome=rec.contig,
                            start=rec.start,
                            end=rec.end,
                            strand=strand,
                            name=rec_name
                        )
                    )

                if len(interval_target) != 0 and rec is not None:
                    self.data.append(Transcript(
                        chromosome=rec.contig,
                        start=start,
                        end=end,
                        strand=strand,
                        exons=interval_target,
                        transcript_id=interval_label,
                        gene_id=interval_label,
                        gene=interval_label,
                        transcript=interval_label,
                        category="interval"
                    ))

            except NameError:
                raise "Target region was not found, run load before add_bed."

            except OSError:
                raise f"Error found when build index for {interval_file}, please sort file manually"


if __name__ == "__main__":
    pass
